// reiserfs-test.cpp
//
// Copyright (c) 2003, Ingo Weinhold (bonefish@cs.tu-berlin.de)
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// You can alternatively use *this file* under the terms of the the MIT
// license included in this package.

#define _IOSTREAM_H
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "Block.h"
#include "BlockCache.h"
#include "cache.h"
#include "Debug.h"
#include "DirItem.h"
#include "IndirectItem.h"
#include "Item.h"
#include "Iterators.h"
#include "Key.h"
#include "StatItem.h"
#include "Tree.h"
#include "VNode.h"
#include "Volume.h"

/*!
	\file reiserfs-test.cpp
	\brief Userland test code.

	I originally started the development with this file. I used it to analyze
	whether I understood the structures and algorithms correctly.
	The program features a little shell to navigate in the file system and
	print information.
	The code is not really up to date with the latest class changes, that is
	it does a lot of things a more complicated way than it was necessary, but
	on the other hand this helped me a lot, when I hit a bug in the file
	system add-on and could cross-check it here.
*/

// min and max
// We don't want to include <algobase.h> otherwise we also get <iostream.h>
// and other undesired things.
template<typename C>
static inline C min(const C &a, const C &b) { return (a < b ? a : b); }
template<typename C>
static inline C max(const C &a, const C &b) { return (a > b ? a : b); }

// dummies for user-land testing
int	get_vnode(nspace_id nsid, vnode_id vnid, void **data) { return B_ERROR; }
int	put_vnode(nspace_id nsid, vnode_id vnid) { return B_ERROR; }
int	new_vnode(nspace_id nsid, vnode_id vnid, void *data) { return B_OK; }
int init_cache_for_device(int fd, off_t max_blocks) { return B_OK; }
int remove_cached_device_blocks(int dev, int allow_write) { return B_OK; }
void *get_block(int dev, off_t bnum, int bsize) { return NULL; }
int release_block(int dev, off_t bnum) { return B_ERROR; }


// dump_super_block
void
dump_super_block(const reiserfs_super_block &sb)
{
	printf("super block:\n");
	printf("blocks:             %lu\n", sb.s_block_count);
	printf("free blocks:        %lu\n", sb.s_free_blocks);
	printf("root block:         %lu (%lx)\n", sb.s_root_block, sb.s_root_block);
	printf("journal block:      %lu\n", sb.s_journal_block);
	printf("journal dev:        %lu\n", sb.s_journal_dev);
	printf("journal size:       %lu\n", sb.s_orig_journal_size);
	printf("journal...\n");
	printf("block size:         %hu\n", sb.s_blocksize);
	printf("oid maxsize:        %hu\n", sb.s_oid_maxsize);
	printf("oid cursize:        %hu\n", sb.s_oid_cursize);
	printf("state:              %hu\n", sb.s_state);
	printf("magic:              `%s'\n", sb.s_magic);
	printf("hash function code: %lu\n", sb.s_hash_function_code);
	printf("tree height:        %u\n", sb.s_tree_height);
	printf("s_bmap_nr:          %hu\n", sb.s_bmap_nr);
	printf("version:            %hu\n", sb.s_version);
	printf("s_inode_generation: %lu\n", sb.s_inode_generation);
}

// print_dir_item
void
print_dir_item(const DirItem &item)
{
	printf("dir item\n");
	for (int32 i = 0; i < item.GetEntryCount(); i++) {
		DirEntry *entry = item.EntryAt(i);
		size_t nameLen = 0;
		const char *name = item.EntryNameAt(i, &nameLen);
		printf("  dir entry:   `%.*s'\n", (int)nameLen, name);
		printf("    offset:    %lu\n", entry->GetOffset());
		printf("    dir ID:    %lu\n", entry->GetDirID());
		printf("    object ID: %lu\n", entry->GetObjectID());
		printf("    location:  %hu\n", entry->GetLocation());
		printf("    state:     %hx\n", entry->GetState());
	}
}

// print_stat_item
void
print_stat_item(Block *block, ItemHeader *itemHead)
{
	if (itemHead->GetType() == TYPE_STAT_DATA) {
		if (itemHead->GetLen() == sizeof(stat_data)) {
			stat_data *data = (stat_data*)((uint8*)block->GetData()
				+ itemHead->GetLocation());
			StatData(data).Dump();
		} else if (itemHead->GetLen() == sizeof(stat_data_v1)) {
			stat_data_v1 *data = (stat_data_v1*)((uint8*)block->GetData()
				+ itemHead->GetLocation());
			StatData(data).Dump();
		} else {
			printf("bad stat data\n");
		}
	}
}

// print_item
void
print_item(Item &item)
{
	ItemHeader *itemHead = item.GetHeader();
	printf("    len:      %hu\n", itemHead->GetLen());
	printf("    location: %hu\n", itemHead->GetLocation());
	printf("    version:  %hu\n", itemHead->GetVersion());
	printf("    type:     %hx\n", itemHead->GetType());
	VKey itemKey;
	printf("    offset:   %Lu\n", item.GetKey(&itemKey)->GetOffset());
	switch (itemHead->GetType()) {
		case TYPE_STAT_DATA:
		{
			StatItem *statItem = static_cast<StatItem*>(&item);
			StatData statData;
			if (statItem->GetStatData(&statData))
				statData.Dump();
			break;
		}
		case TYPE_INDIRECT:
			break;
		case TYPE_DIRECT:
			break;
		case TYPE_DIRENTRY:
		{
			DirItem *dirItem = static_cast<DirItem*>(&item);
			print_dir_item(*dirItem);
			break;
		}
//		case TYPE_ANY:
		default:
			printf("  unknown item type: %hu\n",
				   itemHead->GetKey()->GetType(itemHead->GetVersion()));
			break;
	}
}

// print_leaf_node
void
print_leaf_node(LeafNode *node)
{
	for (int32 i = 0; i < node->CountItems(); i++) {
		Item item;
		if (item.SetTo(node, i) == B_OK) {
			printf("  item %3ld:\n", i);
			print_item(item);
		}
	}
}

// print_node
void
print_node(Node *node)
{
	if (node) {
		printf("block:        %Lu\n", node->GetNumber());
		printf("  level:      %hu\n", node->GetLevel());
		printf("  item count: %hu\n", node->CountItems());
		printf("  free space: %hu\n", node->GetFreeSpace());
		printf("  leaf:       %d\n", node->IsLeaf());
//printf("data: %p\n", node->Data());
//printf("keys: %p\n", node->Keys());
//printf("childs: %p\n", node->Childs());
		if (node->IsInternal()) {
			InternalNode *internal = node->ToInternalNode();
			for (int32 i = 0; i <= internal->CountItems(); i++) {
				const DiskChild *child = internal->ChildAt(i);
				printf("  child %3ld:\n", i);
				printf("    block:   %lu\n", child->GetBlockNumber());
				printf("    size:   %u\n", child->GetSize());
			}
		} else if (node->IsLeaf())
			print_leaf_node(node->ToLeafNode());
	}
}

// get_stat_data
status_t
get_stat_data(Tree *tree, uint32 dirID, uint32 objectID, StatData *statData)
{
//printf("get_stat_data(%lu, %lu)\n", dirID, objectID);
	status_t error = (tree && statData ? B_OK : B_BAD_VALUE);
	// find the stat item
	StatItem item;
	if (error == B_OK) {
		ItemIterator iterator(tree);
		error = iterator.InitCheck();
		VKey k(dirID, objectID, SD_OFFSET, V1_SD_UNIQUENESS, KEY_FORMAT_3_5);
		if (error == B_OK)
			error = iterator.FindRightMost(&k, &item);
	}
	// get the stat data
	if (error == B_OK)
		error = item.GetStatData(statData);
//printf("get_stat_data() done\n");
	return error;
}

// print_file
void
print_file(Tree *tree, uint32 dirID, uint32 objectID, const char *name,
		   bool infoOnly)
{
	Volume *volume = tree->GetVolume();
	VNode node(dirID, objectID);
	if (volume->FindVNode(dirID, objectID, &node) == B_OK) {
//printf("entry: mode: %hx\n", node.GetStatData()->GetMode());
//printf("file?: %d\n", node.IsFile());
		char buffer[4096];
		const size_t bufferSize = sizeof(buffer);
		off_t offset = 0;
		size_t bytesRead;
		while (volume->ReadObject(&node, offset, buffer, bufferSize,
								  &bytesRead) == B_OK
			   && bytesRead > 0) {
//printf("bytes read: %lu\n", bytesRead);
			fwrite(buffer, 1, bytesRead, stdout);
			offset += bytesRead;
		}
		printf("\nEOF\n");
	} else
		printf("Failed to find VNode for entry `%s'\n", name);
}

// entry_type_for_mode
char
entry_type_for_mode(uint16 mode)
{
	if (S_ISREG(mode))
		return ' ';
	if (S_ISLNK(mode))
		return '@';
	if (S_ISDIR(mode))
		return '/';
	return '?';
}

// list_dir
void
list_dir(Tree *tree, uint32 dirID, uint32 objectID)
{
	DirEntryIterator iterator(tree, dirID, objectID);
	DirItem dirItem;
	int32 entryIndex = 0;
	while (iterator.GetNext(&dirItem, &entryIndex) == B_OK) {
		DirEntry *entry = dirItem.EntryAt(entryIndex);
		uint32 entryDirID = entry->GetDirID();
		uint32 entryObjectID = entry->GetObjectID();
//		StatData statData;
VNode node;
//		if (get_stat_data(tree, entryDirID, entryObjectID, &statData)
//			== B_OK) {
if (tree->GetVolume()->FindVNode(entryDirID, entryObjectID, &node) == B_OK) {
StatData &statData = *node.GetStatData();
			size_t nameLen = 0;
			const char *name = dirItem.EntryNameAt(entryIndex, &nameLen);
			printf("%10llu %.*s%c", statData.GetSize(), (int)nameLen, name,
				   entry_type_for_mode(statData.GetMode()));
bool maxInfo = true;
			if (maxInfo) {
				for (int32 i = 0; i < 32 - (int32)nameLen; i++)
					printf(" ");
				printf("(%10lu, %10lu, %10lu)", entryDirID, entryObjectID,
					   entry->GetOffset());
			}
			printf("\n");
		} else
			printf("failed to get stat data for entry...\n");
	}
}

// name_info
struct name_info {
	name_info(const char *name, size_t len) : name(name), len(len) {}
	name_info(int) : name(NULL), len(0) {}	// dummy for list item supplier
	const char	*name;
	size_t		len;
};

typedef List<name_info> NameList;

// print_name_list
void
print_name_list(const NameList *list)
{
	printf(".");
	if (list->CountItems() > 0) {
		for (int32 i = 0; i < list->CountItems(); i++) {
			name_info info = list->ItemAt(i);
			printf("/%.*s", (int)info.len, info.name);
		}
	}
}

// check_dir
void
check_dir(Tree *tree, uint32 dirID, uint32 objectID, NameList *nameList,
		  uint32 level = 0)
{
	DirEntryIterator iterator(tree, dirID, objectID);
	DirItem dirItem;
	int32 entryIndex = 0;
	while (iterator.GetNext(&dirItem, &entryIndex) == B_OK) {
		DirEntry *entry = dirItem.EntryAt(entryIndex);
		uint32 entryDirID = entry->GetDirID();
		uint32 entryObjectID = entry->GetObjectID();
		// get name
		size_t nameLen = 0;
		const char *name = dirItem.EntryNameAt(entryIndex, &nameLen);
		nameList->AddItem(name_info(name, nameLen));
		// check, whether we can find the entry
		DirItem checkDirItem;
		int32 checkEntryIndex = 0;
		if (tree->FindDirEntry(dirID, objectID, name, nameLen,
			&checkDirItem, &checkEntryIndex) == B_OK) {
			if (checkDirItem.GetNode() != dirItem.GetNode()) {
				printf("found entry `");
				print_name_list(nameList);
				printf("' on wrong node!\n");
			} else if (checkDirItem.GetHeader() != dirItem.GetHeader()) {
				printf("found entry `");
				print_name_list(nameList);
				printf("' in wrong item!\n");
			} else if (checkEntryIndex != entryIndex) {
				printf("found entry `");
				print_name_list(nameList);
				printf("' with wrong index!\n");
			}
		} else {
			printf("could find entry `");
			print_name_list(nameList);
			printf("'!\n");
		}
		// get stat and recurse
		StatData statData;
		if (get_stat_data(tree, entryDirID, entryObjectID, &statData)
			== B_OK) {
			if (S_ISDIR(statData.GetMode())
				&& !(nameLen == 1 && name[0] == '.')
				&& !(nameLen == 2 && name[0] == '.' && name[1] == '.')) {
//				for (int32 i = 0; i < (int32)level; i++)
//					printf(" ");
//				printf("%.*s\n", (int)nameLen, name);
				check_dir(tree, entryDirID, entryObjectID, nameList,
						  level + 1);
			}
		} else {
			printf("failed to get stat data for entry `");
			print_name_list(nameList);
			printf("'!\n");
		}
		nameList->RemoveItem(nameList->CountItems() - 1);
	}
}


// shell
void
shell(Tree *tree)
{
	Volume *volume = tree->GetVolume();
	const size_t lineLength = 1024;
	const char *prompt = "> ";
	bool terminate = false;
	char line[lineLength];
	uint32 currentDirDirID = REISERFS_ROOT_PARENT_OBJECTID;
	uint32 currentDirObjectID = REISERFS_ROOT_OBJECTID;
	VNode *currentDir = volume->GetRootVNode();
printf("currentDir->GetStatData()->GetMode(): %hu\n", currentDir->GetStatData()->GetMode());

	while (!terminate) {
		printf(prompt); fflush(stdout);
		terminate = !fgets(line, lineLength, stdin);
		line[lineLength - 1] = '\0';
		// chop line break
		int32 len = strlen(line);
		if (len > 0 && line[len - 1] == '\n') {
			line[len - 1] = '\0';
			len--;
		}
		char command[lineLength];
		if (!terminate && len > 0
			&& sscanf(line, "%s", command) && strlen(command) > 0) {
			const char *parameters = line + strlen(command);
			while (*parameters == ' ')
				parameters++;
			// c
			if (!strcmp(command, "c")) {
				NameList nameList;
				check_dir(tree, currentDirDirID, currentDirObjectID,
						  &nameList);
			// cd
			} else if (!strcmp(command, "cd")) {
				DirItem dirItem;
				int32 entryIndex = 0;
				if (tree->FindDirEntry(currentDirDirID, currentDirObjectID,
						parameters, &dirItem, &entryIndex) == B_OK) {
					DirEntry *entry = dirItem.EntryAt(entryIndex);
					uint32 dirID = entry->GetDirID();
					uint32 objectID = entry->GetObjectID();
					StatData statData;
					if (get_stat_data(tree, dirID, objectID, &statData)
						== B_OK) {
						if (S_ISDIR(statData.GetMode())) {
							currentDirDirID = entry->GetDirID();
							currentDirObjectID = entry->GetObjectID();
						} else
							printf("`%s' is no directory\n", parameters);
					} else {
						printf("failed to get stat for entry `%s'\n",
							   parameters);
					}
				} else
					printf("entry `%s' not found\n", parameters);
			// ll
			} else if (!strcmp(command, "ll") || !strcmp(command, "l")) {
				list_dir(tree, currentDirDirID, currentDirObjectID);
			// p, pi
			} else if (!strcmp(command, "p") || !strcmp(command, "pi")) {
				bool infoOnly = !strcmp(command, "pi");
				DirItem dirItem;
				int32 entryIndex = 0;
				if (tree->FindDirEntry(currentDirDirID, currentDirObjectID,
						parameters, &dirItem, &entryIndex) == B_OK) {
					DirEntry *entry = dirItem.EntryAt(entryIndex);
					uint32 dirID = entry->GetDirID();
					uint32 objectID = entry->GetObjectID();
					print_file(tree, dirID, objectID, parameters, infoOnly);
				} else
					printf("entry `%s' not found\n", parameters);
			// s
			} else if (!strcmp(command, "s")) {
				DirItem dirItem;
				int32 entryIndex = 0;
				if (tree->FindDirEntry(currentDirDirID, currentDirObjectID,
						parameters, &dirItem, &entryIndex) == B_OK) {
					DirEntry *entry = dirItem.EntryAt(entryIndex);
					uint32 dirID = entry->GetDirID();
					uint32 objectID = entry->GetObjectID();
					StatData statData;
					if (get_stat_data(tree, dirID, objectID, &statData)
						== B_OK) {
						statData.Dump();
					} else {
						printf("failed to get stat for entry `%s'\n",
							   parameters);
					}
				} else
					printf("entry `%s' not found\n", parameters);
			} else if (!strcmp(command, "q") || !strcmp(command, "exit")
					   || !strcmp(command, "quit")) {
				terminate = true;
			} else
				printf("unknown command: `%s'\n", command);
		}
	}
	printf("\n");
}


// main
int
main()
{
	init_debugging();
	const char *diskName = "/dev/disk/ide/ata/0/master/0/0_5";
	Volume volume;
	status_t error = volume.Mount(0, diskName);
	if (error == B_OK)
		shell(volume.GetTree());
	else
		printf("failed to init volume: %s\n", strerror(error));
	exit_debugging();
	return 0;
}

